#!/usr/bin/env python
# In case of poor (Sh**y) commenting contact christopher.edelmaier@colorado.edu
# Basic
import sys, os, pdb
import gc
import argparse
import fnmatch
## Analysis
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl

sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'Lib'))
sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'Spindle'))

from spindle_sim import SpindleSim
from sim_graph_funcs import *
from stylelib.ase1_styles import ase1_sims_stl

from scipy.stats import ks_2samp
from scipy.spatial.distance import euclidean
from fastdtw import fastdtw
import scipy.io as sio

'''
Name: BadSpindleAssembly.py
Description: Takes in a good WT plot, and a bad plot, and then does
the spindle length, interpolar fraction, and all attachment states
average over time for them.
Input:
Output:
'''

def parse_args():
    parser = argparse.ArgumentParser(prog='MotorDeleteGraph.py')
    # General options that are actually required
    parser.add_argument('-w', '--wildtype', required=True, type=str,
            help='WT sim')
    parser.add_argument('--bad', required=True, type=str,
            help='Bad sim')
    parser.add_argument('--simname', required=True, type=str,
            help='Simulation Name')

    # Minimum of extra options to make this work
    parser.add_argument('--nopost', action='store_true',
            help="Do not use post analyis program to decrease time of analysis.")
    parser.add_argument('-F', '--fitness', action='store_true',
            help='Create fitness for spindle simulations.')
    parser.add_argument('-A', '--analyze', action='store_true',
            help='Analyze data from multiple simulations')
    parser.add_argument('--datadir', type=str,
            help='Name of the data directory in which all analyzed data files will be read/written. \
                    Also the directory where all the graphs will be placed when saved. \
                    Default is set to {workdir}/data/.')
    parser.add_argument('-d', '--workdir', type=str,
            help='Name of the working directory where simulation will be run.')

    opts = parser.parse_args()
    return opts


# Class definition
class BadSpindleAssembly(object):
    def __init__(self, opts):
        self.opts = opts
        self.cwd = os.getcwd()

        print "opts: {}".format(opts)

        self.ReadOpts()
        self.AnalyzeSims()

    def ReadOpts(self):
        if not self.opts.workdir:
            self.opts.workdir = os.path.abspath(self.cwd)
        elif not os.path.exists(self.opts.workdir):
            raise IOError( "Working directory {} does not exist.".format(
                self.opts.workdir) )
        else:
            self.opts.workdir = os.path.abspath(self.opts.workdir)

        self.wt_dir = os.path.abspath(self.opts.wildtype)
        self.wt_sim = SpindleSim(self.wt_dir, opts)
        self.bad_dir = os.path.abspath(self.opts.bad)
        self.bad_sim = SpindleSim(self.bad_dir, opts)

    def AnalyzeSims(self):
        self.wt_sim.Analyze()
        self.bad_sim.Analyze()
        self.wt_sim.CalcSimSuccess()
        self.bad_sim.CalcSimSuccess()
        if self.opts.fitness:
            wt_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'Data', 'wt.mat')
            lstream_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'Data', 'lstream.mat')
            self.wt_sim.Fitness(wt_path, lstream_path)

    # Graph the avg spindle length for various delete strains (or just multiple strains)
    def GraphSpindleLength(self):
        fig, ax = plt.subplots() 

        nsims = 2
        colors = mpl.cm.rainbow(np.linspace(0,1,nsims))

        # Graph the specific versions of the input data averages
        self.GraphSpindleLengthSim(fig, ax, self.wt_sim, colors, 0, r'WT')
        self.GraphSpindleLengthSim(fig, ax, self.bad_sim, colors, 1, self.opts.simname)

        ax.set_xlabel(r'Time (min)')
        ax.set_ylabel(r'Avg. SPB Separation ($\mu$m)')
        ax.legend()
        fig.tight_layout() 
        plt.savefig('spindle_length_comparison.pdf', dpi=fig.dpi)

    def GraphSpindleLengthSim(self, fig, ax, sim, colors, num, label):
        min_size = 0
        for sd in sim.seeds:
            if (min_size == 0 or min_size > sd.time.size):
                min_size = sd.time.size

        #avg = np.zeros(min_size)
        avg_std = np.zeros((len(sim.seeds), min_size))
        time = sim.seeds[-1].time[:min_size]
        num_seeds = 0

        isd = 0
        for sd in sim.seeds:
            time = sd.PostAnalysis.timedata['time']
            spbsep_dict = sd.PostAnalysis.timedata['spb_separation']
            spbsep_arr = [spbsep_dict[ts] for ts in time ]
            spbsep_arr = np.array(spbsep_arr)*uc['um'][1]
            #avg = np.add(avg, spbsep_arr[:min_size])
            avg_std[isd,:] = spbsep_arr[:min_size]
            num_seeds += 1
            isd += 1

        #print avg_std
        #print avg_std.shape
        #avg /= float(num_seeds)
        #ax.plot(time, avg, color=colors[num], label=label)
        avg = np.mean(avg_std, axis=0)
        avg_stddev = np.std(avg_std, axis=0, ddof=1) / np.sqrt(num_seeds)
        #print avg
        #print avg_stddev
        ax.errorbar(x = time, y = avg, yerr = avg_stddev, color = colors[num], label = label)

    # Graph the avg spindle length for various delete strains (or just multiple strains)
    def GraphSpindleIPF(self):
        fig, ax = plt.subplots() 

        nsims = 2
        colors = mpl.cm.rainbow(np.linspace(0,1,nsims))

        # Graph the specific versions of the input data averages
        self.GraphSpindleIPFSim(fig, ax, self.wt_sim, colors, 0, r'WT')
        self.GraphSpindleIPFSim(fig, ax, self.bad_sim, colors, 1, self.opts.simname)

        ax.set_xlabel(r'Time (min)')
        ax.set_ylabel(r'Avg. Interpolar Fraction')
        ax.legend()
        fig.tight_layout() 
        plt.savefig('spindle_ipf_comparison.pdf', dpi=fig.dpi)

    def GraphSpindleIPFSim(self, fig, ax, sim, colors, num, label):
        min_size = 0
        for sd in sim.seeds:
            if (min_size == 0 or min_size > sd.time.size):
                min_size = sd.time.size

        avg = np.zeros(min_size)
        time = sim.seeds[-1].time[:min_size]
        num_seeds = 0

        for sd in sim.seeds:
            time = sd.PostAnalysis.timedata['time']
            ipf_dict = sd.PostAnalysis.timedata['interpolar_fraction']
            ipf_arr = [ipf_dict[ts] for ts in time ]
            ipf_arr = np.array(ipf_arr)
            avg = np.add(avg, ipf_arr[:min_size])
            num_seeds += 1

        avg /= float(num_seeds)
        ax.plot(time, avg, color=colors[num], label=label)

    # Graph the merotelic attachment state for the simulation over time
    def GraphSpindleAttachStates(self):
        attachment_types = ['unattached',
                            'monotelic',
                            'merotelic',
                            'syntelic',
                            'amphitelic']
        for atype in attachment_types:
            fig, ax = plt.subplots()

            nsims = 2
            colors = mpl.cm.rainbow(np.linspace(0,1,nsims))

            # Graph the specific versions of the input data averages
            self.GraphSpindleStateSim(fig, ax, self.wt_sim, colors, 0, r'WT', atype) 
            self.GraphSpindleStateSim(fig, ax, self.bad_sim, colors, 1, self.opts.simname, atype)

            ax.set_xlabel(r'Time (min)')
            ax.set_ylabel('Avg. N {}'.format(atype))
            ax.legend()
            fig.tight_layout() 
            plt.savefig('spindle_{}_comparison.pdf'.format(atype), dpi=fig.dpi)
            plt.close()

    def GraphSpindleStateSim(self, fig, ax, sim, colors, num, label, atype):
        min_size = 0
        for sd in sim.seeds:
            if (min_size == 0 or min_size > sd.time.size):
                min_size = sd.time.size

        avg = np.zeros(min_size)
        time = sim.seeds[-1].time[:min_size]
        num_seeds = 0

        for sd in sim.seeds:
            #print sd.succ_info_dict
            time = sd.PostAnalysis.timedata['time']
            attach_dict = sd.PostAnalysis.timedata['kc_atypes']
            attach_arr = [attach_dict[ts] for ts in time]
            nchromo = len(attach_arr[0])
            #atype = 'merotelic'
            if atype == 'unattached':
                atypeint = 0
            elif atype == 'monotelic':
                atypeint = 1
            elif atype == 'merotelic':
                atypeint = 2
            elif atype == 'syntelic':
                atypeint = 3
            elif atype == 'amphitelic':
                atypeint = 4

            plot_arr = np.zeros(len(attach_arr))
            
            # Regenerate the merotelic numbers
            for x in xrange(len(attach_arr)):
                for ic in xrange(nchromo):
                    attach = attach_arr[x][ic]
                    if attach == atypeint:
                        plot_arr[x] += 1.0
            avg = np.add(avg, plot_arr[:min_size])
            num_seeds += 1

        avg /= float(num_seeds)
        ax.plot(time, avg, color=colors[num], label=label)

    # Graph the spindle occupany of KC-MT binding sites
    def GraphOccupancyBindingSites():
        fig, ax = plt.subplots()

        nsims = 2
        colors = mpl.cm.rainbow(np.linspace(0,1,nsims))

        # Graph the specific versions of the input data averages
        self.GraphOccupancySim(fig, ax, self.wt_sim, colors, 0, r'WT')
        self.GraphOccupancySim(fig, ax, self.bad_sim, colors, 1, self.opts.simname)

        ax.set_xlabel(r'Time (min)')
        ax.set_ylabel('Avg. N {}'.format(atype))
        ax.legend()
        fig.tight_layout() 
        plt.savefig('spindle_occupancy_comparison.pdf', dpi=fig.dpi)
        plt.close()

    def GraphOccupancySim(self, fig, ax, sim, colors, num, label):
        min_size = 0
        for sd in sim.seeds:
            if (min_size == 0 or min_size > sd.time.size):
                min_size = sd.time.size

        avg = np.zeros(min_size)
        time = sim.seeds[-1].time[:min_size]
        num_seeds = 0

        for sd in sim.seeds:
            #print sd.succ_info_dict
            time = sd.PostAnalysis.timedata['time']
            occupancy_dict = sd.PostAnalysis.timedata['kc_occupancy']
            occupancy_arr = np.array([occupancy_dict[ts] for ts in time])
            avg = np.add(avg, occupancy_arr[:min_size])
            num_seeds += 1

        avg /= float(num_seeds)

################
if __name__ == "__main__":
    opts = parse_args()
    x = BadSpindleAssembly(opts)
    x.GraphSpindleLength()
    x.GraphSpindleIPF()
    x.GraphSpindleAttachStates()
    x.GraphOccupancyBindingSites()
